/* Copyright (C) 2004 - 2005  db4objects Inc.  http://www.db4o.com

This file is part of the db4o open source object database.

db4o is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published
by the Free Software Foundation and as clarified by db4objects' GPL 
interpretation policy, available at
http://www.db4o.com/about/company/legalpolicies/gplinterpretation/
Alternatively you can write to db4objects, Inc., 1900 S Norfolk Street,
Suite 350, San Mateo, CA 94403, USA.

db4o is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */
package com.db4o.nativequery.expr.cmp;

import com.db4o.foundation.*;

// FIXME need to carry more info, must know about Integer.class vs. Integer.TYPE

public class FieldValue implements ComparisonOperand {
	public static interface Visitor extends ComparisonOperandVisitor {

		void visit(FieldValue operand);

	}

	private int _parentIdx;
	private Collection4 _fieldNames = new Collection4();

	public FieldValue(int parentIdx,String name) {
		_parentIdx=parentIdx;
		descend(name);
	}
	
	public FieldValue(int parentIdx,String[] fieldNames) {
		_parentIdx=parentIdx;
		_fieldNames.addAll(fieldNames);
	}

	public FieldValue(int parentIdx, Iterator4 fieldNames) {
		_parentIdx=parentIdx;
		_fieldNames.addAll(fieldNames);
	}

	public FieldValue descend(String fieldName) {
		_fieldNames.add(fieldName);
		return this;
	}

	public Iterator4 fieldNames() {
		return _fieldNames.strictIterator();
	}

	public int parentIdx() {
		return _parentIdx;
	}
	
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (other == null || getClass() != other.getClass()) {
			return false;
		}
		FieldValue casted = (FieldValue) other;
		
		// TODO: implement Collection4#equals to give by-value semantics
		if(_fieldNames.size()!=casted._fieldNames.size()) {
			return false;
		}
		Iterator4 firstIter=_fieldNames.iterator();
		Iterator4 secondIter=casted._fieldNames.iterator();
		while(firstIter.hasNext()) {
			if(!firstIter.next().equals(secondIter.next())) {
				return false;
			}
		}
		return _parentIdx==casted._parentIdx;
	}
	
	public int hashCode() {
		// TODO: implement Collection4#hashCode to give by-value semantics
		int hashCode=0;
		Iterator4 firstIter=_fieldNames.iterator();
		while(firstIter.hasNext()) {
			hashCode*=29+firstIter.next().hashCode();
		}
		return hashCode*29+_parentIdx;
	}
	
	public String toString() {
		StringBuffer str=new StringBuffer();
		str.append(_parentIdx);
		for (Iterator4 nameIter = fieldNames(); nameIter.hasNext();) {
			String fieldName = (String) nameIter.next();
			str.append('.');
			str.append(fieldName);
		}
		return str.toString();
	}
	
	public void accept(ComparisonOperandVisitor visitor) {
		((Visitor)visitor).visit(this);
	}
}
